==> mod_search_rast-0.0.1.tar.gz: mod_search_rast-0.0.1/mc_template.c <==
/* Copyright (C) 2005 Hiroyuki OYAMA
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "mc_template.h"
#include "apr_file_info.h"
#include "apr_file_io.h"
#include "apr_strings.h"

#include <stdio.h>

enum {
    MC_TEMPLATE_PLAIN,
    MC_TEMPLATE_PRINT,
    MC_TEMPLATE_FOR,
    MC_TEMPLATE_IF,
    MC_TEMPLATE_ELSE,
    MC_TEMPLATE_END,
};

typedef struct mc_template_result {
    const char *str;
    struct mc_template_result *next;
} mc_template_result;

static const unsigned char test_char_table[256] = {
    0,62,62,62,62,62,62,62,62,62,63,62,62,62,62,62,62,62,62,62,
    62,62,62,62,62,62,62,62,62,62,62,62,14,0,23,6,1,38,1,1,
    9,9,1,0,8,0,0,10,0,0,0,0,0,0,0,0,0,0,40,15,
    15,8,15,15,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,15,31,15,7,0,7,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,15,39,15,1,62,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54
};
#define TEST_CHAR(c, f)        (test_char_table[(unsigned)(c)] & (f))
#define T_ESCAPE_PATH_SEGMENT  (2)

static const char c2x_table[] = "0123456789abcdef";

static unsigned char *c2x(unsigned what, unsigned char *where)
{
#if APR_CHARSET_EBCDIC
    what = apr_xlate_conv_byte(ap_hdrs_to_ascii, (unsigned char)what);
#endif /*APR_CHARSET_EBCDIC*/
    *where++ = '%';
    *where++ = c2x_table[what >> 4];
    *where++ = c2x_table[what & 0xf];
    return where;
}

static const char *encode_uri(apr_pool_t *p, const char *src)
{
    char *copy = apr_palloc(p, 3 * strlen(src) + 1);
    const unsigned char *s = (const unsigned char *)src;
    unsigned char *d = (unsigned char *)copy;
    unsigned c;

    while ((c = *s)) {
        if (TEST_CHAR(c, T_ESCAPE_PATH_SEGMENT)) {
            d = c2x(c, d);
        }
        else {
            *d++ = c;
        }
        ++s;
    }
    *d = '\0';
    return copy;
}


static const char *encode_html(apr_pool_t *p, const char *s)
{
    int i, j;
    char *x;

    for (i = 0, j = 0; s[i] != '\0'; i++)
        if (s[i] == '<' || s[i] == '>')
            j += 3;
        else if (s[i] == '&')
            j += 4;

    if (j == 0)
        return apr_pstrmemdup(p, s, i);

    x = apr_palloc(p, i + j + 1);
    for (i = 0, j = 0; s[i] != '\0'; i++, j++)
        if (s[i] == '<') {
            memcpy(&x[j], "&lt;", 4);
            j += 3;
        }
        else if (s[i] == '>') {
            memcpy(&x[j], "&gt;", 4);
            j += 3;
        }
        else if (s[i] == '&') {
            memcpy(&x[j], "&amp;", 5);
            j += 4;
        }
        else
            x[j] = s[i];

    x[j] = '\0';
    return (const char *)x;
}

static const char *encode_value(apr_pool_t *p, const char *encode, const char *src)
{
    if (strcmp(encode, "html") == 0) {
        return encode_html(p, src);
    }
    else if (strcmp(encode, "uri") == 0) {
        return encode_uri(p, src);
    }
    else {
        return src;
    }
}


static const char *load_file_body(apr_pool_t *p, const char *path)
{
    char *buff;
    apr_finfo_t finfo;
    apr_size_t len;
    apr_file_t *f = NULL;
    apr_status_t rv;

    rv = apr_stat(&finfo, path, APR_FINFO_PROT, p);
    if (rv != APR_SUCCESS) {
        return NULL;
    }
    len = finfo.size;
    rv = apr_file_open(&f, path, APR_READ, APR_OS_DEFAULT, p);
    if (rv != APR_SUCCESS) {
        return NULL;
    }
    buff = apr_palloc(p, len + 1);

    rv = apr_file_read(f, buff, &len);
    if (rv != APR_SUCCESS) {
        apr_file_close(f);
        return NULL;
    }
    buff[len] = '\0';
    apr_file_close(f);

    return (const char *)buff;
}

static mc_template_node_t *push_node(apr_pool_t *p, mc_template_node_t *dist)
{
    mc_template_node_t *new;

    new = apr_palloc(p, sizeof(mc_template_node_t));
    new->type = MC_TEMPLATE_PLAIN;
    new->arg1 = NULL;
    new->arg2 = NULL;
    new->next = NULL;
    dist->next = new;
    return  new;
}

static int get_print_arg(char *src, char **arg1, char **arg2)
{
    char *pos1, *pos2, *pos3;
    char *attr1, *attr2;
    char *name;

    attr1 = apr_strtok(src, " ", &pos1);
    attr2 = pos1;
    if (attr1 == NULL || attr2 == NULL)
        return 0;

    name = apr_strtok(attr1, "=", &pos2);
    if (name == NULL || strcmp(name, "name") != 0)
        return 0;
    *arg1 = pos2;

    name = apr_strtok(attr2, "=", &pos3);
    if (name == NULL || strcmp(name, "encode") != 0)
        return 0;
    *arg2 = pos3;

    return 1;
}

static int get_name_arg(char *src, char **arg)
{
    char *pos;
    char *name;

    name = apr_strtok(src, "=", &pos);
    if (name == NULL || strcmp(name, "name") != 0)
        return 0;
    *arg = pos;
    return 1;

}

static mc_template_t *tokenize_template(apr_pool_t *p, const char *src)
{
    mc_template_t *tmpl;
    mc_template_node_t *node, *head, *n_node;
    char *cur = (char *)src;
    char *match;
    char *token;


    node = apr_palloc(p, sizeof(mc_template_node_t));
    node->type = MC_TEMPLATE_PLAIN;
    node->arg1 = "";
    node->next = NULL;
    head = node;
    for (;;) {
        match = strstr(cur, "<$");
        if (match == NULL)
            break;
        if (match - cur > 0) {
            token = apr_pstrndup(p, cur, match - cur);
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_PLAIN;
            n_node->arg1 = token;
            node = n_node;
        }
        cur = match + 2;

        match = strstr(cur, "$>");
        if (match == NULL)
            break;
        token = apr_pstrndup(p, cur, match - cur);
        cur = match + 2;

        if (strncmp(token, "print ", strlen("print ")) == 0) {
            char *name, *arg1;
            if (!get_print_arg(token + strlen("print "), &name, &arg1))
                continue;
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_PRINT;
            n_node->name = name;
            n_node->arg1 = arg1;
            node = n_node;
        }
        else if (strncmp(token, "for ", strlen("for ")) == 0) {
            char *name;
            if (!get_name_arg(token + strlen("for "), &name))
                continue;
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_FOR;
            n_node->name = name;
            node = n_node;
        }
        else if (strncmp(token, "if ", strlen("if ")) == 0) {
            char *name;
            if (!get_name_arg(token + strlen("if "), &name))
                continue;
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_IF;
            n_node->name = name;
            node = n_node;
        }
        else if (strncmp(token, "else", strlen("else")) == 0) {
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_ELSE;
            node = n_node;
        }
        else if (strncmp(token, "end", strlen("end")) == 0) {
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_END;
            node = n_node;
        }
        else {
        
        }
    }
    if (*cur != '\0') {
        token = (char *)apr_pstrdup(p, cur);
        n_node = push_node(p, node);
        n_node->type = MC_TEMPLATE_PLAIN;
        n_node->arg1 = token;
        node = n_node;
    }

    tmpl = apr_palloc(p, sizeof(mc_template_t));
    tmpl->src = src;
    tmpl->nodes = head;

    return tmpl;
}

mc_template_t *mc_template_create(apr_pool_t *p, const char *path)
{
    const char *body;

    body = load_file_body(p, path);
    return tokenize_template(p, body);
}

mc_template_t *mc_template_create_by_string(apr_pool_t *p, const char *tmpl)
{
    return tokenize_template(p, tmpl);
}

static mc_template_node_t *skip_block_node(mc_template_node_t *node)
{
    int lc = 0;
    node = node->next;
    for (; node != NULL; node = node->next) {
        if (node->type == MC_TEMPLATE_FOR || node->type == MC_TEMPLATE_IF)
            lc++;
        if (node->type == MC_TEMPLATE_ELSE || node->type == MC_TEMPLATE_END) {
            if (lc > 0) {
                lc--;
                continue;
            }
            break;
        }
    }
    return node;
}

static mc_template_node_t *skip_for_node(mc_template_node_t *node)
{
    int lc = 0;
    node = node->next;
    for (; node != NULL; node = node->next) {
        if (node->type == MC_TEMPLATE_FOR || node->type == MC_TEMPLATE_IF)
            lc++;
        if (node->type == MC_TEMPLATE_END) {
            if (lc > 0) {
                lc--;
                continue;
            }
            break;
        }
    }
    return node;
}

static mc_template_result *exec_template_nodes(apr_pool_t *p,
                                               mc_template_node_t **src_node,
                                               apr_table_t *param)
{
    mc_template_result *result, *head, *new;
    apr_table_t *cur_param;
    mc_template_node_t *node;
    cur_param = param;

    result = apr_palloc(p, sizeof(mc_template_result));
    result->str = "";
    result->next = NULL;
    head = result;
    for (node = *src_node; node != NULL; node = node->next) {
        if (node->type == MC_TEMPLATE_PLAIN) {
            new = apr_palloc(p, sizeof(mc_template_result));
            new->str = node->arg1;
            new->next = NULL;
            result->next = new;
            result = new; 
        }
        else if (node->type == MC_TEMPLATE_PRINT) {
            const char *value;
//            printf("print: name='%s' encode='%s'\n", node->name, node->arg1);
            new = apr_palloc(p, sizeof(mc_template_result));
            value = apr_table_get(cur_param, node->name);
            value = encode_value(p, node->arg1, value ? value : "");
            new->str = value;
            new->next = NULL;
            result->next = new;
            result = new; 
        }
        else if (node->type == MC_TEMPLATE_FOR) {
            apr_array_header_t *list;
            mc_template_node_t *n_head, *n_tail = node->next;
            int i;

            list = (apr_array_header_t *)apr_table_get(cur_param, node->name);
            if (list == NULL) {
                node = skip_for_node(node);
                if (node == NULL)
                    break;
                continue;
            }
            for (i = 0; list != NULL && i < list->nelts; i++) {
                apr_table_t *l_param;
                n_head = node->next;

                l_param = ((apr_table_t **)list->elts)[i];
                l_param = apr_table_overlay(p, l_param, cur_param);
                result->next = exec_template_nodes(p, &n_head, l_param);
                for (;result->next != NULL; result = result->next)
                    ;
                n_tail = n_head;
                if (n_head == NULL)
                    break;
            }
            node = n_tail;
            if (node == NULL)
                break;
        }
        else if (node->type == MC_TEMPLATE_IF) {
            const char *cond;
//            printf("if: '%s'\n", node->name);
            cond = apr_table_get(cur_param, node->name);
            if (cond != NULL) {
                node = node->next;
                result->next = exec_template_nodes(p, &node, cur_param);
                if (node != NULL && node->type != MC_TEMPLATE_END)
                    node = skip_block_node(node);
                for (;result->next != NULL; result = result->next)
                    ;
                if (node == NULL)
                    break;
            }
            else {
                node = node->next;
                node = skip_block_node(node);
                if (node == NULL)
                    break;
                if (node->type == MC_TEMPLATE_END) {
                    continue;
                }
                node = node->next;
                result->next = exec_template_nodes(p, &node, cur_param);

                for (;result->next != NULL; result = result->next)
                    ;
                if (node == NULL)
                    break;
            }
        }
        else if (node->type == MC_TEMPLATE_ELSE) {
//            printf("else:\n");
            break;
        }
        else if (node->type == MC_TEMPLATE_END) {
//            printf("end:\n");
            break;
        }
        else {

        }
    }

    *src_node = node;
    return head;
}

const char *mc_template_build(apr_pool_t *p,
                              mc_template_t *tmpl, apr_table_t *param)
{
    mc_template_result *result, *head;
    char *body;
    int length = 0;
    mc_template_node_t *node = tmpl->nodes;

    result = apr_palloc(p, sizeof(mc_template_result));
    result->str = "";
    result->next = NULL;
    head = result;
    result->next = exec_template_nodes(p, &node, param);

    for (result = head; result != NULL; result = result->next) {
        length += strlen(result->str);
    }
    body = apr_palloc(p, length + 1);
    body[0] = '\0';
    for (result = head; result != NULL; result = result->next) {
        strcat(body, result->str);
    }
    body[length] = '\0';

    return (const char *)body;
}

==> mod_search_rast-0.0.2.tar.gz: mod_search_rast-0.0.2/mc_template.c <==
/* Copyright (C) 2005 Hiroyuki OYAMA
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "mc_template.h"
#include "apr_file_info.h"
#include "apr_file_io.h"
#include "apr_strings.h"

#include <stdio.h>

enum {
    MC_TEMPLATE_PLAIN,
    MC_TEMPLATE_PRINT,
    MC_TEMPLATE_FOR,
    MC_TEMPLATE_IF,
    MC_TEMPLATE_ELSE,
    MC_TEMPLATE_END,
};

typedef struct mc_template_result {
    const char *str;
    struct mc_template_result *next;
} mc_template_result;

static const unsigned char test_char_table[256] = {
    0,62,62,62,62,62,62,62,62,62,63,62,62,62,62,62,62,62,62,62,
    62,62,62,62,62,62,62,62,62,62,62,62,14,0,23,6,1,38,1,1,
    9,9,1,0,8,0,0,10,0,0,0,0,0,0,0,0,0,0,40,15,
    15,8,15,15,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,15,31,15,7,0,7,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,15,39,15,1,62,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,
    54,54,54,54,54,54,54,54,54,54,54,54,54,54,54,54
};
#define TEST_CHAR(c, f)        (test_char_table[(unsigned)(c)] & (f))
#define T_ESCAPE_PATH_SEGMENT  (2)

static const char c2x_table[] = "0123456789abcdef";

static unsigned char *c2x(unsigned what, unsigned char *where)
{
#if APR_CHARSET_EBCDIC
    what = apr_xlate_conv_byte(ap_hdrs_to_ascii, (unsigned char)what);
#endif /*APR_CHARSET_EBCDIC*/
    *where++ = '%';
    *where++ = c2x_table[what >> 4];
    *where++ = c2x_table[what & 0xf];
    return where;
}

static const char *encode_uri(apr_pool_t *p, const char *src)
{
    char *copy = apr_palloc(p, 3 * strlen(src) + 1);
    const unsigned char *s = (const unsigned char *)src;
    unsigned char *d = (unsigned char *)copy;
    unsigned c;

    while ((c = *s)) {
        if (TEST_CHAR(c, T_ESCAPE_PATH_SEGMENT)) {
            d = c2x(c, d);
        }
        else {
            *d++ = c;
        }
        ++s;
    }
    *d = '\0';
    return copy;
}


static const char *encode_html(apr_pool_t *p, const char *s)
{
    int i, j;
    char *x;

    for (i = 0, j = 0; s[i] != '\0'; i++)
        if (s[i] == '<' || s[i] == '>')
            j += 3;
        else if (s[i] == '&')
            j += 4;

    if (j == 0)
        return apr_pstrmemdup(p, s, i);

    x = apr_palloc(p, i + j + 1);
    for (i = 0, j = 0; s[i] != '\0'; i++, j++)
        if (s[i] == '<') {
            memcpy(&x[j], "&lt;", 4);
            j += 3;
        }
        else if (s[i] == '>') {
            memcpy(&x[j], "&gt;", 4);
            j += 3;
        }
        else if (s[i] == '&') {
            memcpy(&x[j], "&amp;", 5);
            j += 4;
        }
        else
            x[j] = s[i];

    x[j] = '\0';
    return (const char *)x;
}

static const char *encode_value(apr_pool_t *p, const char *encode, const char *src)
{
    if (strcmp(encode, "html") == 0) {
        return encode_html(p, src);
    }
    else if (strcmp(encode, "uri") == 0) {
        return encode_uri(p, src);
    }
    else {
        return src;
    }
}


static const char *load_file_body(apr_pool_t *p, const char *path)
{
    char *buff;
    apr_finfo_t finfo;
    apr_size_t len;
    apr_file_t *f = NULL;
    apr_status_t rv;

    rv = apr_stat(&finfo, path, APR_FINFO_PROT, p);
    if (rv != APR_SUCCESS) {
        return NULL;
    }
    len = finfo.size;
    rv = apr_file_open(&f, path, APR_READ, APR_OS_DEFAULT, p);
    if (rv != APR_SUCCESS) {
        return NULL;
    }
    buff = apr_palloc(p, len + 1);

    rv = apr_file_read(f, buff, &len);
    if (rv != APR_SUCCESS) {
        apr_file_close(f);
        return NULL;
    }
    buff[len] = '\0';
    apr_file_close(f);

    return (const char *)buff;
}

static mc_template_node_t *push_node(apr_pool_t *p, mc_template_node_t *dist)
{
    mc_template_node_t *new;

    new = apr_palloc(p, sizeof(mc_template_node_t));
    new->type = MC_TEMPLATE_PLAIN;
    new->arg1 = NULL;
    new->arg2 = NULL;
    new->next = NULL;
    dist->next = new;
    return  new;
}

static int get_print_arg(char *src, char **arg1, char **arg2)
{
    char *pos1, *pos2, *pos3;
    char *attr1, *attr2;
    char *name;

    attr1 = apr_strtok(src, " ", &pos1);
    attr2 = pos1;
    if (attr1 == NULL || attr2 == NULL)
        return 0;

    name = apr_strtok(attr1, "=", &pos2);
    if (name == NULL || strcmp(name, "name") != 0)
        return 0;
    *arg1 = pos2;

    name = apr_strtok(attr2, "=", &pos3);
    if (name == NULL || strcmp(name, "encode") != 0)
        return 0;
    *arg2 = pos3;

    return 1;
}

static int get_name_arg(char *src, char **arg)
{
    char *pos;
    char *name;

    name = apr_strtok(src, "=", &pos);
    if (name == NULL || strcmp(name, "name") != 0)
        return 0;
    *arg = pos;
    return 1;

}

static mc_template_t *tokenize_template(apr_pool_t *p, const char *src)
{
    mc_template_t *tmpl;
    mc_template_node_t *node, *head, *n_node;
    char *cur = (char *)src;
    char *match;
    char *token;


    node = apr_palloc(p, sizeof(mc_template_node_t));
    node->type = MC_TEMPLATE_PLAIN;
    node->arg1 = "";
    node->next = NULL;
    head = node;
    for (;;) {
        match = strstr(cur, "<$");
        if (match == NULL)
            break;
        if (match - cur > 0) {
            token = apr_pstrndup(p, cur, match - cur);
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_PLAIN;
            n_node->arg1 = token;
            node = n_node;
        }
        cur = match + 2;

        match = strstr(cur, "$>");
        if (match == NULL)
            break;
        token = apr_pstrndup(p, cur, match - cur);
        cur = match + 2;

        if (strncmp(token, "print ", strlen("print ")) == 0) {
            char *name, *arg1;
            if (!get_print_arg(token + strlen("print "), &name, &arg1))
                continue;
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_PRINT;
            n_node->name = name;
            n_node->arg1 = arg1;
            node = n_node;
        }
        else if (strncmp(token, "for ", strlen("for ")) == 0) {
            char *name;
            if (!get_name_arg(token + strlen("for "), &name))
                continue;
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_FOR;
            n_node->name = name;
            node = n_node;
        }
        else if (strncmp(token, "if ", strlen("if ")) == 0) {
            char *name;
            if (!get_name_arg(token + strlen("if "), &name))
                continue;
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_IF;
            n_node->name = name;
            node = n_node;
        }
        else if (strncmp(token, "else", strlen("else")) == 0) {
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_ELSE;
            node = n_node;
        }
        else if (strncmp(token, "end", strlen("end")) == 0) {
            n_node = push_node(p, node);
            n_node->type = MC_TEMPLATE_END;
            node = n_node;
        }
        else {
        
        }
    }
    if (*cur != '\0') {
        token = (char *)apr_pstrdup(p, cur);
        n_node = push_node(p, node);
        n_node->type = MC_TEMPLATE_PLAIN;
        n_node->arg1 = token;
        node = n_node;
    }

    tmpl = apr_palloc(p, sizeof(mc_template_t));
    tmpl->src = src;
    tmpl->nodes = head;

    return tmpl;
}

mc_template_t *mc_template_create(apr_pool_t *p, const char *path)
{
    const char *body;
    mc_template_t *t;

    body = load_file_body(p, path);
    if (body == NULL)
        return NULL;
    t = tokenize_template(p, body);
    if (t != NULL)
        t->content_type = NULL;
    return t;
}

mc_template_t *mc_template_create_by_string(apr_pool_t *p, const char *tmpl)
{
    if (tmpl == NULL)
        return NULL;
    return tokenize_template(p, tmpl);
}

static mc_template_node_t *skip_block_node(mc_template_node_t *node)
{
    int lc = 0;
    node = node->next;
    for (; node != NULL; node = node->next) {
        if (node->type == MC_TEMPLATE_FOR || node->type == MC_TEMPLATE_IF)
            lc++;
        if (node->type == MC_TEMPLATE_ELSE || node->type == MC_TEMPLATE_END) {
            if (lc > 0) {
                lc--;
                continue;
            }
            break;
        }
    }
    return node;
}

static mc_template_node_t *skip_for_node(mc_template_node_t *node)
{
    int lc = 0;
    node = node->next;
    for (; node != NULL; node = node->next) {
        if (node->type == MC_TEMPLATE_FOR || node->type == MC_TEMPLATE_IF)
            lc++;
        if (node->type == MC_TEMPLATE_END) {
            if (lc > 0) {
                lc--;
                continue;
            }
            break;
        }
    }
    return node;
}

static mc_template_result *exec_template_nodes(apr_pool_t *p,
                                               mc_template_node_t **src_node,
                                               apr_table_t *param)
{
    mc_template_result *result, *head, *new;
    apr_table_t *cur_param;
    mc_template_node_t *node;
    cur_param = param;

    result = apr_palloc(p, sizeof(mc_template_result));
    result->str = "";
    result->next = NULL;
    head = result;
    for (node = *src_node; node != NULL; node = node->next) {
        if (node->type == MC_TEMPLATE_PLAIN) {
            new = apr_palloc(p, sizeof(mc_template_result));
            new->str = node->arg1;
            new->next = NULL;
            result->next = new;
            result = new; 
        }
        else if (node->type == MC_TEMPLATE_PRINT) {
            const char *value;
//            printf("print: name='%s' encode='%s'\n", node->name, node->arg1);
            new = apr_palloc(p, sizeof(mc_template_result));
            value = apr_table_get(cur_param, node->name);
            value = encode_value(p, node->arg1, value ? value : "");
            new->str = value;
            new->next = NULL;
            result->next = new;
            result = new; 
        }
        else if (node->type == MC_TEMPLATE_FOR) {
            apr_array_header_t *list;
            mc_template_node_t *n_head, *n_tail = node->next;
            int i;

            list = (apr_array_header_t *)apr_table_get(cur_param, node->name);
            if (list == NULL) {
                node = skip_for_node(node);
                if (node == NULL)
                    break;
                continue;
            }
            for (i = 0; list != NULL && i < list->nelts; i++) {
                apr_table_t *l_param;
                n_head = node->next;

                l_param = ((apr_table_t **)list->elts)[i];
                l_param = apr_table_overlay(p, l_param, cur_param);
                result->next = exec_template_nodes(p, &n_head, l_param);
                for (;result->next != NULL; result = result->next)
                    ;
                n_tail = n_head;
                if (n_head == NULL)
                    break;
            }
            node = n_tail;
            if (node == NULL)
                break;
        }
        else if (node->type == MC_TEMPLATE_IF) {
            const char *cond;
//            printf("if: '%s'\n", node->name);
            cond = apr_table_get(cur_param, node->name);
            if (cond != NULL) {
                node = node->next;
                result->next = exec_template_nodes(p, &node, cur_param);
                if (node != NULL && node->type != MC_TEMPLATE_END)
                    node = skip_block_node(node);
                for (;result->next != NULL; result = result->next)
                    ;
                if (node == NULL)
                    break;
            }
            else {
                node = node->next;
                node = skip_block_node(node);
                if (node == NULL)
                    break;
                if (node->type == MC_TEMPLATE_END) {
                    continue;
                }
                node = node->next;
                result->next = exec_template_nodes(p, &node, cur_param);

                for (;result->next != NULL; result = result->next)
                    ;
                if (node == NULL)
                    break;
            }
        }
        else if (node->type == MC_TEMPLATE_ELSE) {
//            printf("else:\n");
            break;
        }
        else if (node->type == MC_TEMPLATE_END) {
//            printf("end:\n");
            break;
        }
        else {

        }
    }

    *src_node = node;
    return head;
}

const char *mc_template_build(apr_pool_t *p,
                              mc_template_t *tmpl, apr_table_t *param)
{
    mc_template_result *result, *head;
    char *body;
    int length = 0;
    mc_template_node_t *node = tmpl->nodes;

    result = apr_palloc(p, sizeof(mc_template_result));
    result->str = "";
    result->next = NULL;
    head = result;
    result->next = exec_template_nodes(p, &node, param);

    for (result = head; result != NULL; result = result->next) {
        length += strlen(result->str);
    }
    body = apr_palloc(p, length + 1);
    body[0] = '\0';
    for (result = head; result != NULL; result = result->next) {
        strcat(body, result->str);
    }
    body[length] = '\0';

    return (const char *)body;
}
